/*____________________________________________________________________________

	Implementation of XGIFPicture class
____________________________________________________________________________*/
#include "stdafx.h"
#include <sys/types.h>
#include <sys/stat.h>
#include "GifPict.hpp"


/*____________________________________________________________________________

								  Prototypes
____________________________________________________________________________*/
short GetByte(void);
XeStatus OutLine(unsigned char *pixels, short sLineLen);
XeStatus decoder( short linewidth );


/*____________________________________________________________________________

								  Globals
____________________________________________________________________________*/
static	unsigned char *	pTheGIF;
static	unsigned char *	pTheGIFStart;
static	unsigned char *	pTheBaseAddr;
static	unsigned char *	pTheBaseAddrStart;
static	short	sTheRowBytes, sThePixelSize, sTheHeight;
static	long	lTheDataSize;

short	bad_code_count;

// Static variables
static short curr_size;                     // The current code size
static short clear;                         // Value for a clear code
static short ending;                        // Value for a ending code
static short newcodes;                      // First available code
static short top_slot;                      // Highest code for current size
static short slot;                          // Last read code

// The following static variables are used for seperating out codes
static short navail_bytes = 0;              // # bytes left in block
static short nbits_left = 0;                // # bits left in current byte
static unsigned char b1;                    // Current byte
static unsigned char byte_buff[257];        // Current block
static unsigned char *pbytes;               // Pointer to next byte in block

static long code_mask[13] = {

	 0,
	 0x0001, 0x0003,
	 0x0007, 0x000F,
	 0x001F, 0x003F,
	 0x007F, 0x00FF,
	 0x01FF, 0x03FF,
	 0x07FF, 0x0FFF
};

static short init_exp(short size);
static short get_next_code(void);


/*_________________________________________________________________________________

	Notes:	The reason we have these seperated like this instead of using
			a structure like the original Wilhite code did, is because this
			stuff generally produces significantly faster code when compiled...
			This code is full of similar speedups...  (For a good book on writing
			C for speed or for space optomisation, see Efficient C by Tom Plum,
			published by Plum-Hall Associates...)
_________________________________________________________________________________*/

static unsigned char stack[MAX_CODES + 1];            // Stack for storing pixels
static unsigned char suffix[MAX_CODES + 1];           // Suffix table
static unsigned short prefix[MAX_CODES + 1];           // Prefix linked list






/*__________________________________________________________________________________
__________________________________________________________________________________*/
inline short WIDTHBYTES(BITMAPINFOHEADER * psDibHeader)
{
	return (((psDibHeader->biWidth * psDibHeader->biBitCount) + 31) / 32 * 4);
}





/*__________________________________________________________________________________

__________________________________________________________________________________*/
BOOL IsGIFHeader(const char * pGIF)
{
	static const char xkGIF_HEADER_STRING [] = "GIF87a";

	if (strncmp(xkGIF_HEADER_STRING, pGIF, strlen(xkGIF_HEADER_STRING)) == 0)
		return TRUE;
	else
		return FALSE;
}



/*__________________________________________________________________________________

	parseGIFHeader()

	Notes:	This routine parses the gif header and returns an error if the file
			is not GIF87a.
__________________________________________________________________________________*/

XeStatus 
XGIFPicture::ParseGIFHeader(unsigned char * pGIF, unsigned char * & beginningOfGIFImage)
{
	unsigned char * pRunning = pGIF;

	if ( ! IsGIFHeader((const char *) pRunning) )
		return xeFAILURE;

	pRunning += 6;		// skip past GIF87a string
#if 0
	pRunning += 4; 		// skip past screen width & height
#else
	munWidth = * (short *) pRunning;
	pRunning += 2;
	munHeight = * (short *) pRunning;
	pRunning += 2;
#endif

	unsigned char cFlag = *(pRunning++);
	unsigned short bGlobalColorMap = cFlag & 0x80;

	munBitsPerPixel	= (cFlag & 0x07) + 1;

	// DANGER - Hack ahead (I'm sure there is a better way to do this...)
	if(munBitsPerPixel > 0 && munBitsPerPixel < 2)
		munBitsPerPixel = 2;								
	if(munBitsPerPixel > 2 && munBitsPerPixel < 4)
		munBitsPerPixel = 4;
	if(munBitsPerPixel > 4 && munBitsPerPixel < 8)
		munBitsPerPixel = 8;
	if(munBitsPerPixel > 8 && munBitsPerPixel < 16)
		munBitsPerPixel = 16;
	if(munBitsPerPixel > 16 && munBitsPerPixel < 32)
		munBitsPerPixel = 32;

	pRunning++; 		// skip background color
	pRunning++;  		// skip zero byte

	// This is the number of colors in the color table
	unsigned short usColorMapSize;
	if (bGlobalColorMap )
		usColorMapSize = (1 << munBitsPerPixel);
	else
		usColorMapSize = 0;


	// parse the global color map
	// This is total bytes needed for DIB header and DIB color table
	unsigned long ulDIBHeaderSize = 
				  sizeof(BITMAPINFOHEADER)
				+ (usColorMapSize * sizeof(RGBQUAD));

	// Allocate DIB header and color table		
	mpsDibHeader = (BITMAPINFO *) new char [ulDIBHeaderSize];

	// Initialize DIB color table
	RGBQUAD * rgbPtr = &mpsDibHeader->bmiColors[0];
	RGBQUAD * tempPtr = &mpsDibHeader->bmiColors[0];
	for (int i = 0; i < usColorMapSize; i++)
	{		
		rgbPtr->rgbRed   = (*(pRunning++));	
		rgbPtr->rgbGreen = (*(pRunning++));
		rgbPtr->rgbBlue  = (*(pRunning++));
		rgbPtr->rgbReserved = 0;
		rgbPtr++;
	}		

	pRunning += 5;   // skip past byteValue, leftPicOffset, topPicOffset
	
	////	munWidth	= gpsToolbox->swapWord(* (short *) pRunning, XToolbox::xeINTEL_BYTE_ORDER);
///	munWidth = * (short *) pRunning;
	pRunning += 2;
	
	////	munHeight	= gpsToolbox->swapWord(* (short *) pRunning, XToolbox::xeINTEL_BYTE_ORDER);
//	munHeight = * (short *) pRunning;
	pRunning += 2;
	
	pRunning++;		// skip 2nd gif flag 

	// Initialize DIB header
	mpsDibHeader->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	mpsDibHeader->bmiHeader.biWidth = munWidth;
	// Negating sign here makes the normally upside-down DIB format draw right-side up.
	mpsDibHeader->bmiHeader.biHeight = -munHeight;
	mpsDibHeader->bmiHeader.biPlanes = 1;
	mpsDibHeader->bmiHeader.biBitCount = munBitsPerPixel;
	mpsDibHeader->bmiHeader.biCompression = BI_RGB;
	mpsDibHeader->bmiHeader.biSizeImage = 0;
	mpsDibHeader->bmiHeader.biXPelsPerMeter = 0;
	mpsDibHeader->bmiHeader.biYPelsPerMeter = 0;
	mpsDibHeader->bmiHeader.biClrUsed = 0;
	mpsDibHeader->bmiHeader.biClrImportant = 0;

	mpPictureData = new unsigned char[WIDTHBYTES(&mpsDibHeader->bmiHeader) * munHeight];

	beginningOfGIFImage = pRunning;		// Return current position in data where image starts

	return xeSUCCESS;
}



/*__________________________________________________________________________________

	DrawGIF()
	
	Notes:	This routine draws the gif in the rect.
__________________________________________________________________________________*/

XeStatus DrawGIF(	short sWidth, 
					short sHeight, 
					unsigned char * pRawImage,// Raw GIF image
					BITMAPINFO *	psDibHeader,	// Header and palette
					unsigned char * pPictureData,	// pointer to the DIB pixel data
					long lDataSize )
{
	
	pTheGIF				= pRawImage;
	pTheGIFStart		= pRawImage;
	pTheBaseAddrStart	= pPictureData;
	pTheBaseAddr		= pPictureData;
	sTheRowBytes		= WIDTHBYTES(&psDibHeader->bmiHeader);
	sThePixelSize		= psDibHeader->bmiHeader.biBitCount;
	sTheHeight			= sHeight;
	lTheDataSize		= lDataSize;
	
	return decoder(sWidth);
}


/*__________________________________________________________________________________

	GetByte()

	Notes:	This external (machine specific) function is expected to return
			either the next byte from the GIF file, or a negative number.
__________________________________________________________________________________*/

short GetByte(void){

	if (lTheDataSize > 0L)
	{
		if (pTheGIF < pTheGIFStart + lTheDataSize)
			return *(pTheGIF++) & 0x00FF;
		else
		{
			return -1;
		}
	}
	else
		return *(pTheGIF++) & 0x00FF;
}


/*__________________________________________________________________________________

	OutLine()
	
	Notes:	This function takes a full line of pixels (one byte per pixel) and
			displays them (or does whatever your program wants with them...).  It
			should return zero, or negative if an error or some other event occurs
			which would require aborting the decode process...  Note that the length
			passed will almost always be equal to the line length passed to the
			decoder function, with the sole exception occurring when an ending code
			occurs in an odd place in the GIF file...  In any case, linelen will be
			equal to the number of pixels passed...
__________________________________________________________________________________*/

XeStatus OutLine(unsigned char *pixels, short sLineLen)
{
	short	i, j;
	char	cByte;
///	long	lOffset;
	
	if (pTheBaseAddr >= (pTheBaseAddrStart + ((long) sTheRowBytes * (long) sTheHeight)))
	{
		return xeFAILURE;
	}
	
	switch (sThePixelSize)
	{
		case 1 :
			for (i=0; i<sLineLen; i+=8)
			{
				cByte = 0;
				
				for (j=0; j<8; j++)
					if (pixels[i+j] > 0)
						cByte += (1<<(7-j));
						
				pTheBaseAddr[(i>>3)] = cByte;
			}
			
			break;
			
		case 4 :
			for (i=0; i<sLineLen; i+=2)
				pTheBaseAddr[(i>>1)] = ((pixels[i] << 4) + pixels[i + 1]);
				
			break;
			
		case 8 :
			for (i=0; i<sLineLen; i++)
				pTheBaseAddr[i] = pixels[i];
				
			break;
	}
	
	pTheBaseAddr += (long) sTheRowBytes;
	
	return xeSUCCESS;
}


/*_________________________________________________________________________________

	init_exp()
	
	Notes:	This function initializes the decoder for reading a new image.
_________________________________________________________________________________*/

static short init_exp(short size){

	curr_size = size + 1;
	top_slot = 1 << curr_size;
	clear = 1 << size;
	ending = clear + 1;
	slot = newcodes = ending + 1;
	navail_bytes = nbits_left = 0;
	
	return(0);
}

/*_________________________________________________________________________________

	get_next_code()
	
	Notes:	This routine gets the next code from the GIF file.  Returns the code, 
	or else a negative number in case of file errors...
_________________________________________________________________________________*/

static short get_next_code(void){

	short i, x;
	unsigned long ret;

	if (nbits_left == 0){
	
		if (navail_bytes <= 0){

			// Out of bytes in current block, so read next block
			pbytes = byte_buff;
			
			if ((navail_bytes = GetByte()) < 0)
				return(navail_bytes);

			else if (navail_bytes){
			
				for (i = 0; i < navail_bytes; ++i)
				{
					if ((x = GetByte()) < 0)
						return(x);
						
					byte_buff[i] = (unsigned char) x;
				}
			}
		}
		
		b1 = *pbytes++;
		nbits_left = 8;
		--navail_bytes;
	}

	ret = b1 >> (8 - nbits_left);
	while (curr_size > nbits_left){
	
		if (navail_bytes <= 0){
		
			// Out of bytes in current block, so read next block
			pbytes = byte_buff;
			if ((navail_bytes = GetByte()) < 0)
				return(navail_bytes);
				
			else if (navail_bytes)
			{
				for (i = 0; i < navail_bytes; ++i)
				{
					if ((x = GetByte()) < 0)
						return x;
						
					byte_buff[i] = (unsigned char) x;
				}
			}
		}
		
		b1 = *pbytes++;
		ret |= b1 << nbits_left;
		nbits_left += 8;
		--navail_bytes;
	}

	nbits_left -= curr_size;
	ret &= code_mask[curr_size];
	
	return((short)(ret));
}


/*_________________________________________________________________________________

	decoder()

	Notes:	This function decodes an LZW image, according to the method used
			in the GIF spec.  Every *linewidth* "characters" (ie. pixels) decoded
			will generate a call to OutLine(), which is a user specific function
			to display a line of pixels.  The function gets it's codes from
			get_next_code() which is responsible for reading blocks of data and
			seperating them into the proper size codes.  Finally, GetByte() is
			the global routine to read the next byte from the GIF file.

			It is generally a good idea to have linewidth correspond to the actual
			width of a line (as specified in the Image header) to make your own
			code a bit simpler, but it isn't absolutely necessary.
	
			Returns: 0 if successful, else negative.  (See ERRS.H)
_________________________________________________________________________________*/

XeStatus decoder(short linewidth)
{
	register unsigned char *sp, *bufptr;
	unsigned char *buf;
	short code, fc, oc, bufcnt;
	short c, size;
	XeStatus ret;

	// Initialize for decoding a new image...
	if ((size = GetByte()) < 0)
		return xeFAILURE;
      
	if (size < 2 || 9 < size)
		return xeFILE_CORRUPT;
		
	init_exp(size);

	// Initialize in case they forgot to put in a clear code.
	oc = fc = 0;

	// Allocate space for the decode buffer
	if ((buf = new unsigned char[linewidth + 1]) == NULL)
		return xeMEMORY;

	// Set up the stack pointer and decode buffer pointer
	sp = stack;
	bufptr = buf;
	bufcnt = linewidth;

/*_________________________________________________________________________________
   
	Notes:	This is the main loop.  For each code we get we pass through the
			linked list of prefix codes, pushing the corresponding "character" for
			each code onto the stack.  When the list reaches a single "character"
			we push that on the stack too, and then start unstacking each
			character for output in the correct order.  Special handling is
			included for the clear code, and the whole thing ends when we get
			an ending code.
_________________________________________________________________________________*/

	while ((c = get_next_code()) != ending){

		// If we had a file error, return without completing the decode
		if (c < 0)
		{
			delete [] buf;
			return xeFAILURE;
		}

		// If the code is a clear code, reinitialize all necessary items.
		if (c == clear){
		
			curr_size = size + 1;
			slot = newcodes;
			top_slot = 1 << curr_size;

			// Continue reading codes until we get a non-clear code
			while ((c = get_next_code()) == clear)
				;
	
			// If we get an ending code immediately after a clear code
			if (c == ending)
				break;

/*_________________________________________________________________________________

	Notes:	Finally, if the code is beyond the range of already set codes,
			(This one had better NOT happen...  I have no idea what will
			result from this, but I doubt it will look good...) then set it
			to color zero.
_________________________________________________________________________________*/

			if (c >= slot)
				c = 0;
	
			oc = fc = c;

/*_________________________________________________________________________________

	Notes:	And let us not forget to put the char into the buffer... And
			if, on the off chance, we were exactly one pixel from the end
			of the line, we have to send the buffer to the OutLine()
			routine...
_________________________________________________________________________________*/

			*bufptr++ = (unsigned char) c;
			if (--bufcnt == 0)
			{			
				if ((ret = OutLine(buf, linewidth)) != xeSUCCESS)
				{
					delete [] buf;
					return ret;
				}
				
				bufptr = buf;
				bufcnt = linewidth;
			}
		
		}else{

/*_________________________________________________________________________________

	Notes:	In this case, it's not a clear code or an ending code, so
			it must be a code code...  So we can now decode the code into
			a stack of character codes. (Clear as mud, right?)
_________________________________________________________________________________*/

         	code = c;

/*_________________________________________________________________________________

	Notes:	Here we go again with one of those off chances...  If, on the
			off chance, the code we got is beyond the range of those already
			set up (Another thing which had better NOT happen...) we trick
			the decoder into thinking it actually got the last code read.
			(Hmmn... I'm not sure why this works...  But it does...)
_________________________________________________________________________________*/

			if (code >= slot){
			
				if (code > slot)
					++bad_code_count;
	               
				code = oc;
				*sp++ = (unsigned char) fc;
			}

/*_________________________________________________________________________________

	Notes:	Here we scan back along the linked list of prefixes, pushing
			helpless characters (ie. suffixes) onto the stack as we do so.
_________________________________________________________________________________*/

			while (code >= newcodes){
			
				*sp++ = suffix[code];
				code = prefix[code];
			}

/*_________________________________________________________________________________

	Notes:	Push the last character on the stack, and set up the new
			prefix and suffix, and if the required slot number is greater
			than that allowed by the current bit size, increase the bit
			size.  (NOTE - If we are all full, we *don't* save the new
			suffix and prefix...  I'm not certain if this is correct...
			it might be more proper to overwrite the last code...
_________________________________________________________________________________*/

			*sp++ = (unsigned char) code;
			if (slot < top_slot)
			{
				fc = code;
				suffix[slot] = (unsigned char) fc;
				prefix[slot++] = oc;
				oc = c;
			}
		
			if (slot >= top_slot){
			
				if (curr_size < 12){
				
					top_slot <<= 1;
					++curr_size;
				}
			}

/*_________________________________________________________________________________

	Notes:	Now that we've pushed the decoded string (in reverse order)
			onto the stack, lets pop it off and put it into our decode
			buffer...  And when the decode buffer is full, write another
			line...
_________________________________________________________________________________*/

			while (sp > stack){
			
				*bufptr++ = *(--sp);
				
				if (--bufcnt == 0){
				
					if ((ret = OutLine(buf, linewidth)) != xeSUCCESS)
					{					
						delete [] buf;
						return ret ;
					}
					
					bufptr = buf;
					bufcnt = linewidth;
				}
			}
		}
	}
	
	ret = xeFAILURE;

	if (bufcnt != linewidth)
		ret = OutLine(buf, (linewidth - bufcnt));

	delete [] buf;
	
	return ret;
}





/*____________________________________________________________________________
____________________________________________________________________________*/
XGIFPicture::XGIFPicture(CFile & file) : XPictureBase()
{
	unsigned char * rawImage;

	// Allocate temporary chunk of memory to hold entire GIF file in memory
	if ((rawImage = new unsigned char[file.GetLength()]) != NULL)
	{
		// Read in entire GIF file
		if (file.ReadHuge(rawImage, file.GetLength()) == file.GetLength()) 
		{
			unsigned char * resultantPtr;

			// Chunk that GIF image into a DIB
			if (ParseGIFHeader(rawImage, resultantPtr) == xeSUCCESS)
				mnStatus = loadImage(resultantPtr);	// Read rest of image
			else
				mnStatus = xeFILE_CORRUPT;
		}

		// Get rid of temp memory
		delete [] rawImage;
	}
}



/*____________________________________________________________________________
____________________________________________________________________________*/
XGIFPicture::~XGIFPicture()
{
	if (mpsDibHeader != NULL)
		delete [] (char *) mpsDibHeader;

	if (mpPictureData != NULL)
		delete [] mpPictureData;
}




/*____________________________________________________________________________

	loadImage()
	
	Notes:
____________________________________________________________________________*/

XeStatus XGIFPicture::loadImage(unsigned char * rawImage)
{
	//XRectangle rPict;
	XeStatus err = xeSUCCESS;
///	static GWorldPtr offscreenPort = NULL;
///	CGrafPtr savedPort;
////	GDHandle savedDevice;
////	PixMapHandle hPixMap;
////	CTabHandle hCTable = NULL;
	XeStatus nStatus = xeSUCCESS;		// Assume success; reset otherwise if we fail.

#if 0	
	RECT offscreenRect, rRect;
	RECT oOriginal;
	// calculate the picture rectangle
	rPict = XRectangle(0, 0, width(), height());
	oOriginal = rPict;
	
	
	// create gworld
	GetGWorld(&savedPort, &savedDevice);

	if (mpsDibHeader != NULL)
	{
		hCTable = (CTabHandle) NewHandle(munColorTableSize);
		if (hCTable == NULL)
		{
			xdERROR(TRUE,"could not creat CTable");
			mnStatus = xeMEMORY;
			return mnStatus;
		}
		Palette2CTab(mpsDibHeader, hCTable);
		SetPalette(FrontWindow(), mpsDibHeader, TRUE);
		CTabChanged(hCTable);
	}
	else
		hCTable = NULL;

	offscreenRect = gpsToolbox->convertToRect(oOriginal);
#endif

#if 0	
	// create the offscreen world
	err = NewGWorld(&mpGWorld, munBitsPerPixel, &offscreenRect,hCTable, NULL, 0);
	if (mpGWorld == NULL || err != 0)
	{
		xdERROR(TRUE,"can't create ofscreen world");
		mnStatus = xeMEMORY;
		return mnStatus;
	}
	SetGWorld(mpGWorld, NULL);

	RGBColor rgb;
	
	gpsToolbox->convertRGB(gpsToolbox->getColor(XRgbColor::xeBLACK), &rgb);
	RGBForeColor(&rgb);
	
	gpsToolbox->convertRGB(gpsToolbox->getColor(XRgbColor::xeWHITE), &rgb);
	RGBBackColor(&rgb);

	hPixMap = GetGWorldPixMap(mpGWorld);

	LockPixels(hPixMap);

	rRect = gpsToolbox->convertToRect(rPict);
	EraseRect(&offscreenRect);
#endif

	// decode GIF image
	DrawGIF(	munWidth, 
				munHeight, 
				rawImage,
				mpsDibHeader, 
				mpPictureData,
				//GetPixBaseAddr(hPixMap),
				//(*hPixMap)->rowBytes & 0x1FFF, 
				//(*hPixMap)->pixelSize,
				mlImageSize
			);

#if 0
	SetGWorld(savedPort, NULL);

	UnlockPixels(hPixMap);
#endif

	return nStatus;
}



short 
XGIFPicture::height() const
{
	return munHeight;
}

short 		
XGIFPicture::width() const
{
	return munWidth;
}


XeStatus	
XGIFPicture::draw(CDC & destDC, const RECT & destRect)
{
///	int nPreviousMode = destDC.SetStretchBltMode(CDC::STRETCH_DELETESCANS);
	StretchDIBits(	destDC.GetSafeHdc(), 
					0, 
					0,
					destRect.right, 
					destRect.bottom,
					0,
					0,
					munWidth, 
					munHeight,
					mpPictureData,
					mpsDibHeader,
					DIB_RGB_COLORS,
					SRCCOPY
				);
///	destDC.SetStretchBltMode(nPreviousMode);
		   
	return xeSUCCESS;
}
